/*
 *  $Id$
 *  Copyright (c) 2003, Dominik 'Aeneas' Schnitzer <dominik@schnitzer.at>
 *
 *  MoC - Master of Celebration is an automatic playlist generation tool using
 *  state-of-the-art music similarity measures
 *  http://www.schnitzer.at/dominik/moc/
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <valarray>
#include <iostream>
#include <fstream>
#include <cstdlib>
#include <string>
#include <map>
#include <algorithm>
#include <utility>

#include "vec.h"
#include "mat.h"
#include "gmm.h"
#include "mathtools.h"

using namespace std;

gmm::gmm(const string &fname, const string &md5, vec &prior, mat &mean,
        mat &cov)
    : filename(fname), md5sum(md5), selfdist(-1.0)
{
    int i;

    gaussians = prior.size();
    gaussiandim = mean[0].size();
    priors.resize(gaussians);
    priors = prior;

    means.resize(gaussians);
    covs.resize(gaussians);
    for (i = 0; i < gaussians; i++)
    {
        means[i].resize(gaussiandim);
        covs[i].resize(gaussiandim);
        means[i] = mean[i];
        covs[i] = cov[i];
    }
}

gmm::gmm(const string &fname)
    : selfdist(-1.0)
{
    load_model(fname);
}


gmm::~gmm()
{
}

double
gmm::distance(gmm *g, int datacount)
{
    // lookup if we already computed the distance
    map<gmm*, double>::iterator p = distcache.find(g);
    if (p != distcache.end())
        return p->second;
    
    int row, i, j;
    
    double normal = pow (2.0*M_PI, gaussiandim/2.0);
    vec a(0.0, gaussians), prob(0.0, gaussians);
    vec s(1.0, gaussians);
    vec diff(gaussiandim);
    vec data(gaussiandim);
    int nextrow = 0;

    mat *other_means = g->get_means();
    mat *other_covs = g->get_covs();
    vec *other_priors = g->get_priors();
    mat other_var(gaussians);

    // This mixture parameters against given featdata
    for (i = 0; i < gaussians; i++)
    {
        other_var[i].resize(gaussiandim);
        for (j = 0; j < gaussiandim; j++)
        {
            other_var[i][j] = sqrt((*other_covs)[i][j]);
            s[i] *= covs[i][j];
        }
        s[i] = sqrt(s[i]) * normal;
    }

    int gc = 0;
    nextrow = (int)((*other_priors)[gc]*datacount);

    for (row = 0; row < datacount; row++)
    {
        if (row > nextrow)
        {
            gc++;
            if (gc == (gaussians-1))
                nextrow = datacount;
            else
                nextrow += (int)((*other_priors)[gc]*datacount);
        }

        for (j = 0; j < gaussiandim; j++)
            data[j] = nrnd3((*other_means)[gc][j], other_var[gc][j], row);

        for (i = 0; i < gaussians; i++)
        {
            diff = data - means[i];
            diff *= diff;
            diff /= covs[i];
            a[i] = exp(-0.5*diff.sum());
        }
        a /= s;
        a *= priors;
        prob += a;
    }

    double dist = prob.sum();
    distcache[g] = dist;
    g->set_distance(this, dist);

    // distance
    return dist;
}

void
gmm::set_distance(gmm *g, double dist)
{
    distcache[g] = dist;
}

double
gmm::get_distance(gmm *g, int datacount)
{
    return (abs(1.0 - (distance(g, datacount)/get_selfdistance(datacount))) +
        abs(1.0 - g->distance(this, datacount)/g->get_selfdistance(datacount)))
        / 2.0;
}

double
gmm::get_selfdistance(int datacount)
{
    if (selfdist < 0.0)
        selfdist = distance(this, datacount);

    return selfdist;
}

int
gmm::get_gaussians()
{
    return gaussians;
}

int
gmm::get_gaussiandim()
{
    return gaussiandim;
}

mat*
gmm::get_means()
{
    return &means;
}

mat*
gmm::get_covs()
{
    return &covs;
}

vec*
gmm::get_priors()
{
    return &priors;
}

string
gmm::get_filename()
{
    return filename;
}

string
gmm::get_md5sum()
{
    return md5sum;
}   

void
gmm::print_model(ostream &c)
{
    int i, j;
 
    c.precision(10);
    c << filename << endl;
    c << "GMM: Mixture-Models " << gaussians << " Gauss-Dimensions "
        << gaussiandim << endl;
    c << "MD5: " << md5sum << endl;
    c << "Covariances:" << endl;
    for (i = 0; i < gaussians; i++)
    {
        c << "Model " << i << ": ";
        for (j = 0; j < gaussiandim; j++)
            c << covs[i][j] << " ";
        c << endl;
    }
 
    c << "Means:" << endl;
    for (i = 0; i < gaussians; i++)
    {
        c << "Model " << i << ": ";
        for (j = 0; j < gaussiandim; j++)
            c << means[i][j] << " ";
        c << endl;
    }
 
    c << "Priors:" << endl;
    for (i = 0; i < gaussians; i++)
        c << "Model " << i << ": " << priors[i] << endl;
}

void
gmm::load_model(const string &fname)
{
    string dummy;
    int i, j;
    ifstream inf(fname.c_str());
    char fle[2048];

    inf.getline(fle, 2048);
    filename = fle;
    // GMM: Mixture-Models //gaussians// Gauss-Dimensions //gaussiandim//
    inf >> dummy >> dummy >> gaussians >> dummy >> gaussiandim;
    // MD5: //md5sum//
    inf >> dummy >> md5sum;

    // Covariances:
    inf >> dummy;
    covs.resize(gaussians);
    for (i = 0; i < gaussians; i++)
    {
        covs[i].resize(gaussiandim);
        // Model X: //covs[0][0]// ...
        inf >> dummy >> dummy;
        for (j = 0; j < gaussiandim; j++)
            inf >> covs[i][j];
    }

    // Means:
    inf >> dummy;
    means.resize(gaussians);
    for (i = 0; i < gaussians; i++)
    {
        means[i].resize(gaussiandim);
        // Model X: //means[0][0]// ...
        inf >> dummy >> dummy;
        for (j = 0; j < gaussiandim; j++)
            inf >> means[i][j];
    }

    // Priors:
    inf >> dummy;
    priors.resize(gaussians);
    for (i = 0; i < gaussians; i++)
        inf >> dummy >> dummy >> priors[i];

    inf.close();
}

void
gmm::save_model(const string &fname)
{
    ofstream outf(fname.c_str());
    print_model(outf);
    outf.close();
}

