/* -*- c-basic-offset: 2 -*-
 * GStreamer
 * Copyright (C) 1999-2001 Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <string.h>
#include <gst/gst.h>
#include <gst/audio/audio.h>
#include "gstreplaygain.h"
#include "gain_analysis.h"

#define PASSTHRU_BUF_SIZE 4096
#define PASSTHRU_NUM_BUFS 4

/* elementfactory information */
static GstElementDetails replaygain_details = {
  "Replaygain",
  "Filter/Audio/Effect",
  "LGPL",
  "Calculate the replaygain of a music piece (www.replaygain.org)",
  VERSION,
  "Dominik 'Aeneas' Schnitzer <dominik@schnitzer.at>",
  "(C) 2003",
};

enum {
  /* FILL ME */
  LAST_SIGNAL
};

/* static guint gst_filter_signals[LAST_SIGNAL] = { 0 }; */

enum {
  ARG_0,
  ARG_SILENT,
  ARG_RPGAIN
};

static GstPadTemplate*                    
replaygain_sink_factory (void)            
{                                         
  static GstPadTemplate *template = NULL; 
                                          
  if (! template) {                        
    template = gst_pad_template_new 
      ("sink", GST_PAD_SINK, GST_PAD_ALWAYS,
      gst_caps_append (gst_caps_new ("sink_int",  "audio/raw",
                                     GST_AUDIO_INT_PAD_TEMPLATE_PROPS),
                       gst_caps_new ("sink_float", "audio/raw",
                                     GST_AUDIO_FLOAT_MONO_PAD_TEMPLATE_PROPS)),
      NULL);
  }                                       
  return template;                        
}

static GstPadTemplate*
replaygain_src_factory (void)
{
  static GstPadTemplate *template = NULL;
  
  if (! template)
    template = gst_pad_template_new 
      ("src", GST_PAD_SRC, GST_PAD_ALWAYS,
       gst_caps_append (gst_caps_new ("src_float", "audio/raw",
                                      GST_AUDIO_FLOAT_MONO_PAD_TEMPLATE_PROPS),
                        gst_caps_new ("src_int", "audio/raw",
                                      GST_AUDIO_INT_PAD_TEMPLATE_PROPS)),
       NULL);
  
  return template;
}

static void		replaygain_class_init		(GstReplaygainClass *klass);
static void		replaygain_init		(GstReplaygain *filter);

static void		replaygain_set_property	(GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
static void		replaygain_get_property	(GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);

static GstPadLinkReturn replaygain_connect_sink	(GstPad *pad, GstCaps *caps);

static void		replaygain_chain		(GstPad *pad, GstBuffer *buf);
static void inline 	replaygain_fast_float_chain 	(gfloat* data, guint numsamples, GstReplaygain *filter);
static void inline 	replaygain_fast_16bit_chain 	(gint16* data, guint numsamples);
static void inline 	replaygain_fast_8bit_chain	(gint8* data, guint numsamples);

static GstElementClass *parent_class = NULL;

static GstBufferPool*
replaygain_get_bufferpool (GstPad *pad)
{
  GstReplaygain *filter;

  filter = GST_REPLAYGAIN (gst_pad_get_parent (pad));

  return gst_pad_get_bufferpool (filter->srcpad);
}

static GstPadLinkReturn
replaygain_connect_sink (GstPad *pad, GstCaps *caps)
{
  const gchar *format;
  GstReplaygain *filter;
  
  g_return_val_if_fail (pad  != NULL, GST_PAD_LINK_DELAYED);
  g_return_val_if_fail (caps != NULL, GST_PAD_LINK_DELAYED);

  filter = GST_REPLAYGAIN (gst_pad_get_parent (pad));
  g_return_val_if_fail (filter != NULL, GST_PAD_LINK_REFUSED);
  g_return_val_if_fail (GST_IS_REPLAYGAIN (filter), GST_PAD_LINK_REFUSED);

  gst_caps_get_string(caps, "format", &format);
  
  gst_caps_get_int (caps, "rate", &filter->rate);
  gst_caps_get_int (caps, "channels", &filter->channels);
  
  if (strcmp (format, "int") == 0) {
    filter->format        = GST_REPLAYGAIN_FORMAT_INT;
    gst_caps_get_int	 (caps, "width",      &filter->width);
    gst_caps_get_int	 (caps, "depth",      &filter->depth);
    gst_caps_get_int	 (caps, "law",        &filter->law);
    gst_caps_get_int	 (caps, "endianness", &filter->endianness);
    gst_caps_get_boolean (caps, "signed",     &filter->is_signed);

    if (! filter->silent) {
      g_print ("Replaygain : channels %d, rate %d\n",  
               filter->channels, filter->rate);
      g_print ("Replaygain : format int, bit width %d, endianness %d, signed %s\n",
               filter->width, filter->endianness, filter->is_signed ? "yes" : "no");
    }
  } else if (strcmp (format, "float") == 0) {
    filter->format     = GST_REPLAYGAIN_FORMAT_FLOAT;
    gst_caps_get_string (caps, "layout",    &filter->layout);
    gst_caps_get_float  (caps, "intercept", &filter->intercept);
    gst_caps_get_float  (caps, "slope",     &filter->slope);

    if (! filter->silent) {
      g_print ("Replaygain : channels %d, rate %d\n",  
               filter->channels, filter->rate);
      g_print ("Replaygain : format float, layout %s, intercept %f, slope %f\n",
               filter->layout, filter->intercept, filter->slope);
    }
  }

  

  if (GST_CAPS_IS_FIXED (caps))
  {
    InitGainAnalysis(filter->rate);
    return gst_pad_try_set_caps (filter->srcpad, caps);
  }
  return GST_PAD_LINK_DELAYED;
}

GType
gst_replaygain_get_type (void)
{
  static GType replaygain_type = 0;

  if (!replaygain_type) {
    static const GTypeInfo replaygain_info = {
      sizeof (GstReplaygainClass),      NULL,
      NULL,
      (GClassInitFunc) replaygain_class_init,
      NULL,
      NULL,
      sizeof (GstReplaygain),
      0,
      (GInstanceInitFunc) replaygain_init,
    };
    replaygain_type = g_type_register_static (GST_TYPE_ELEMENT, "GstReplaygain", &replaygain_info, 0);
  }
  return replaygain_type;
}

static void
replaygain_class_init (GstReplaygainClass *klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass*) klass;
  gstelement_class = (GstElementClass*) klass;

  parent_class = g_type_class_ref(GST_TYPE_ELEMENT);

  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_SILENT,
    g_param_spec_boolean("silent","silent","silent",
                         TRUE, G_PARAM_READWRITE));
  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_RPGAIN,
    g_param_spec_double("rpgain","rpgain","rpgain",
                        0.0, G_MAXDOUBLE, 0.0, G_PARAM_READABLE));

  gobject_class->set_property = replaygain_set_property;
  gobject_class->get_property = replaygain_get_property;
}

static void
replaygain_init (GstReplaygain *filter)
{
  filter->srcpad = gst_pad_new_from_template (replaygain_src_factory (),"src");
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  filter->sinkpad = gst_pad_new_from_template (replaygain_sink_factory (),"sink");
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);

  gst_pad_set_link_function    (filter->sinkpad, replaygain_connect_sink);
  gst_pad_set_bufferpool_function (filter->sinkpad, replaygain_get_bufferpool);  
  gst_pad_set_chain_function      (filter->sinkpad, replaygain_chain);

  filter->silent = FALSE;
}

static void
replaygain_chain (GstPad *pad, GstBuffer *buf)
{
  GstReplaygain *filter;
  gint16 *int_data;
  gfloat *float_data;
  
  g_return_if_fail (pad != NULL);
  g_return_if_fail (GST_IS_PAD (pad));
  g_return_if_fail (buf != NULL);
  
  filter = GST_REPLAYGAIN (gst_pad_get_parent (pad));
  g_return_if_fail (filter != NULL);
  g_return_if_fail (GST_IS_REPLAYGAIN (filter));

  filter->bufpool = gst_pad_get_bufferpool (filter->srcpad);
  if (filter->bufpool == NULL) {
    filter->bufpool = gst_buffer_pool_get_default (PASSTHRU_BUF_SIZE, PASSTHRU_NUM_BUFS);
  }
  
  switch (filter->format) {
  case GST_REPLAYGAIN_FORMAT_INT:
    int_data = (gint16 *) GST_BUFFER_DATA (buf);

    switch (filter->width) {
    case 16:
      replaygain_fast_16bit_chain (int_data, GST_BUFFER_SIZE (buf) / 2);
      break;
    case 8:
      replaygain_fast_8bit_chain ((gint8*) int_data, GST_BUFFER_SIZE (buf));
      break;
    }

    break;
  case GST_REPLAYGAIN_FORMAT_FLOAT:
    float_data = (gfloat *) GST_BUFFER_DATA (buf);
    
    replaygain_fast_float_chain (float_data, GST_BUFFER_SIZE (buf) / sizeof (gfloat), filter);
    
    break;
  }
  
  gst_pad_push (filter->srcpad, buf);
}

static void inline
replaygain_fast_float_chain(gfloat* data, guint num_samples, GstReplaygain *filter)
{
  g_print("float chain not implemented use float\n");
}

static void inline
replaygain_fast_16bit_chain(gint16* data, guint num_samples)
{
  Float_t ddata[num_samples];
  int i;
  for (i = 0; i < num_samples; i++)
      ddata[i] = (Float_t)data[i];
  AnalyzeSamples(ddata, NULL, num_samples, 1);
}


static void inline
replaygain_fast_8bit_chain(gint8* data, guint num_samples)
{
  g_print("8bit chain not implemented use float\n");
}


static void
replaygain_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec)
{
  GstReplaygain *filter;

  g_return_if_fail (GST_IS_REPLAYGAIN (object));
  filter = GST_REPLAYGAIN (object);

  switch (prop_id) 
  {
    case ARG_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    default:
      break;
  }
}

static void
replaygain_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec)
{
  GstReplaygain *filter;

  g_return_if_fail (GST_IS_REPLAYGAIN (object));
  filter = GST_REPLAYGAIN (object);

  switch (prop_id) {
    case ARG_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case ARG_RPGAIN:
      g_value_set_double (value, (double)GetTitleGain());
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static gboolean
plugin_init (GModule *module, GstPlugin *plugin)
{
  GstElementFactory *factory;

  factory = gst_element_factory_new ("replaygain", GST_TYPE_REPLAYGAIN, &replaygain_details);
  g_return_val_if_fail (factory != NULL, FALSE);
  
  gst_element_factory_add_pad_template (factory, replaygain_src_factory ());
  gst_element_factory_add_pad_template (factory, replaygain_sink_factory ());

  gst_plugin_add_feature (plugin, GST_PLUGIN_FEATURE (factory));

  return TRUE;
}

GST_PLUGIN_DESC_STATIC(GST_VERSION_MAJOR, GST_VERSION_MINOR,
            "replaygain", plugin_init);
