/*
 *  $Id$
 *  Copyright (c) 2003, Dominik 'Aeneas' Schnitzer <dominik@schnitzer.at>
 *
 *  MoC - Master of Celebration is an automatic playlist generation tool using
 *  state-of-the-art music similarity measures
 *  http://www.schnitzer.at/dominik/moc/
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <valarray>
#include "mfcc.h"
#include <iostream>

using namespace std;

mfcc::mfcc(unsigned int winsize, unsigned int sampr, unsigned int cepcoeffs)
{
    lowestfreq = 133.3333;
    linf = 13;
    logf = 27;
    totf = linf + logf;
    linsp = 66.66666666;
    logsp = 1.0711703;
    max = 0.0;

    lower.resize(totf);
    center.resize(totf);
    upper.resize(totf);
    triangleh.resize(totf);
    freqs.resize(totf + 2);

    earmag.resize(totf);
    calcfb.resize(winsize);
    calcdct.resize(totf);

    this->winsize = winsize;
    this->sampr = sampr;
    this->cepcoeffs = cepcoeffs;

    // Linear filter boundaries = 0:(linf-1)
    for (unsigned int i = 0; i < linf; i++)
        freqs[i] = lowestfreq + i * linsp;

    // Logarithmic filter boundaries = (linf-1):
    for (unsigned int i = 1; i <= (logf+2); i++)
        freqs[linf-1+i] = freqs[linf-1] * pow(logsp, (double)i);

    // Triangles information
    for (unsigned int i = 0; i < totf; i++)
        lower[i] = freqs[i];
    for (unsigned int i = 1; i <= totf; i++)
        center[i-1] = freqs[i];
    for (unsigned int i = 2; i <= totf + 1; i++)
        upper[i-2] = freqs[i];
    for (unsigned int i = 0; i < totf; i++)
        triangleh[i] = 2.0 / (upper[i] - lower[i]);

    // FFT Frequencies
    fftf.resize(winsize);
    for (unsigned int i = 0; i < winsize; i++)
        fftf[i] = (double)i / (double)winsize * (double)sampr;

    // Initialize FilterWeights
    mfccfw.resize(totf);
    for (unsigned int chan = 0; chan < totf; chan++)
    {
        mfccfw[chan].resize(winsize);
        for (unsigned int i = 0; i < winsize; i++)
        {
            mfccfw[chan][i] = 0.0;
            if ((fftf[i] > lower[chan]) && (fftf[i] <= center[chan]))
            {
                mfccfw[chan][i] += triangleh[chan] *
                    ((fftf[i] - lower[chan])/(center[chan] - lower[chan]));
            }
            if ((fftf[i] > center[chan]) && (fftf[i] < upper[chan]))
            {
                mfccfw[chan][i] += triangleh[chan] *
                    ((upper[chan] - fftf[i])/(upper[chan] - center[chan]));
            }
        }
    }

    // Initialize MFCC_DCT
    double scalefac = 1.0 / sqrt(totf/2.0);
    mfccdct.resize(cepcoeffs);
    for (unsigned int j = 0; j < cepcoeffs; j++)
    {
        mfccdct[j].resize(totf);
        for (unsigned int i = 0; i < totf; i++)
            mfccdct[j][i] = scalefac * cos(j * (2*i +1) * M_PIl/2/totf);

        mfccdct[j][0] *= sqrt(2.0)/2.0;
    }
}

mfcc::~mfcc()
{
}

void
mfcc::process(vec *in, vec *out)
{
    // Calculate the filterbank responce
    double sum = 0.0;
    for (unsigned int i = 0; i < totf; i++)
    {
        calcfb = mfccfw[i] * (*in);
        sum = calcfb.sum();
        if (sum != 0.0)
            earmag[i] = log10(sum);
        else
            earmag[i] = 0.0;
    }

    // Take the DCT
    for (unsigned int i = 0; i < cepcoeffs; i++)
    {
        calcdct = mfccdct[i] * earmag;
        (*out)[i] = calcdct.sum();
    }
}

