/*
 *  $Id$
 *  Copyright (c) 2003, Dominik 'Aeneas' Schnitzer <dominik@schnitzer.at>
 *
 *  MoC - Master of Celebration is an automatic playlist generation tool using
 *  state-of-the-art music similarity measures
 *  http://www.schnitzer.at/dominik/moc/
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

using namespace std;

#include <iostream>
#include <fstream>
#include <valarray>
#include <vector>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <unistd.h>
#include <string>
#include <map>

#include "moc.h"
#include "md5.h"
#include "audiofilein.h"
#include "stft.h"
#include "mfcc.h"
#include "gmm.h"
#include "mathtools.h"

// FIXME: ugly constants
#define MIN_COVARIANCE  1.0e-9
#define GMM_WIDTH       1.0

string dbloc = ".moc";

moc::moc()
{
    gaussians = 3;
    window_size = 1024;
    sample_rate = 11025;
    mfcc_coeffs = 8;
    gaussiandim = mfcc_coeffs;
    datacount = 0;
    progr = 0.0;
    cancel = 0;

    // Resize the mean/covariance/variance vectors to fit the data
    mean.resize(gaussians);
    cov.resize(gaussians);
    for (int i = 0; i < gaussians; i++)
    {
        mean[i].resize(gaussiandim);
        cov[i].resize(gaussiandim);
    }
    prior.resize(gaussians);

    init_database();
}

moc::moc(int gaussians, int window_size = 1024, int sample_rate = 11025,
        int mfcc_coeffs = 8)
{
    this->progr = 0.0;
    this->gaussians = gaussians;
    this->window_size = window_size;
    this->sample_rate = sample_rate;
    this->mfcc_coeffs = mfcc_coeffs;
    gaussiandim = mfcc_coeffs;
    datacount = 0;
    cancel = 0;

    // Resize the mean/covariance/variance vectors to fit the data
    mean.resize(gaussians);
    cov.resize(gaussians);
    for (int i = 0; i < gaussians; i++)
    {
        mean[i].resize(gaussiandim);
        cov[i].resize(gaussiandim);
    }
    prior.resize(gaussians);

    init_database();
}

moc::~moc()
{
    typedef map<string, gmm*>::const_iterator ci;
    for (ci p = mdb.begin(); p != mdb.end(); p++)
        delete p->second;
}

void
moc::init_database()
{
    struct stat st;
    DIR *dir;
    struct dirent *dire;
    char *olddir = getcwd(NULL, 0);

    db = string(getenv("HOME")) + "/" + dbloc + "/";
    if (stat(db.c_str(), &st) == -1)
        mkdir(db.c_str(), 0755);

    dir = opendir(db.c_str());
    
    chdir(db.c_str());
    while ((dire = readdir(dir)) != NULL)
    {
        stat(dire->d_name, &st);
        if ((S_ISREG(st.st_mode)) && (strlen(dire->d_name) == 32))
        {
            gmm *g = new gmm(string(dire->d_name));
            mdb[g->get_md5sum()] = g;
        }
    }

    chdir(olddir);
    free(olddir);
}

gmm*
moc::get_gmm(const string &fname, int kmeans, int em)
{
    md5sum = get_md5(fname);
    if (mdb[md5sum] != NULL)
        return mdb[md5sum];

    if (get_mfcc(fname) <= 0)
        return NULL;

    // Compute means and covariances by clustering
    clusterby_kmeans(kmeans);
    clusterby_em(em);

    string filen = realpath(fname.c_str(), 0);
    gmm *g = new gmm(filen, md5sum, prior, mean, cov);

    delete data;
    return g;
}

string
moc::get_md5(const string &fname)
{
    // Compute the md5sum of the file
    int dec = open(fname.c_str(), O_RDONLY);
    md5_t       md5struct;
    short       md5res[16];
    char        md5str[33];
    int         rd;
    short       s[4096];

    md5_init(&md5struct);
    while ((rd = read(dec, &s, sizeof(s))) > 0)
        md5_process(&md5struct, &s, rd);
    close(dec);
    md5_finish(&md5struct, &md5res);
    md5_sig_to_string(&md5res, md5str, 33);

    return string(md5str);
}

int
moc::get_mfcc(const string &fname)
{
    double rps;
    string tempfile = ain.decode(fname, sample_rate, &rps);

    if (tempfile == "")
        return -1;

    int dec = open(tempfile.c_str(), O_RDONLY);

    stft        stft(window_size);
    vec         fftdata(window_size);
    mfcc        mfcc(window_size, sample_rate, mfcc_coeffs);
    vec         mfccdata(mfcc_coeffs);
    float       s[window_size/2];
    double      in[window_size];
    int         wincount = 0;
    int         readpos = 0;
    int         i;
    vector<vec> d;

    // Read the data
    // This loop discards the last (incomplete dataset)
    while (read(dec, &s, sizeof(s)) == (int) sizeof(s))
    {
        // Repleygain normalizing :) www.replaygain.org
        for (i = 0; i < window_size/2; i++)
            in[i+readpos] = (double)(s[i]) * rps; // / 32768.0;

        readpos += window_size/2;
        if (readpos >= window_size)
        {
            stft.process(in, &fftdata);
            mfcc.process(&fftdata, &mfccdata);

            d.push_back(mfccdata);

            // overlap half of the window
            memcpy(&(in[0]), &(in[window_size/2]),
                    window_size/2 * sizeof(double));
            readpos = window_size/2;
            wincount++;
        }
    }

    close(dec);
    unlink(tempfile.c_str());

    datacount = d.size();
    data = new mat(datacount);
    for (i = 0; i < datacount; i++)
    {
        (*data)[i].resize(gaussiandim);
        (*data)[i] = d[i];
    }

    return datacount;
}

void
moc::clusterby_kmeans(int kmeans)
{
    mat old_mean(gaussians);
    vec quality(gaussians);
    vec old_quality(gaussians);
    valarray<int> likelihood(gaussians);
    valarray<int> featclass(datacount);
    double min;
    double dist;
    int i, j, row, iter;

    // Initialize all models with random data/gaussians
    for (i = 0; i < gaussians; i++)
    {
        int randrow = i*(datacount/gaussians) + (int)
            ((double)(rand())/(double)(RAND_MAX) * (datacount/gaussians));
        mean[i] = (*data)[randrow];

        old_mean[i].resize(gaussiandim);
    }
    quality = 0.0;

    // Perform K-Means to initialize the GMM for EM Training
    for (iter = 0; iter < kmeans; iter++)
    {
        // Step 1: Initialization
        for (i = 0; i < gaussians; i++)
        {
            old_mean[i] = mean[i];
            mean[i] = 0.0;
            cov[i] = 0.0;
        }
        old_quality = quality;
        likelihood = 0;

        // Step 2: Search for the nearest data-center
        for (row = 0; row < datacount; row++)
        {
            min = euclidean((*data)[row], old_mean[0]);
            featclass[row] = 0;

            for (i = 1; i < gaussians; i++)
            {
                dist = euclidean((*data)[row], old_mean[i]);
                if (dist < min)
                {
                    featclass[row] = i;
                    min = dist;
                }
            }
            mean[featclass[row]] += (*data)[row];
            likelihood[featclass[row]]++;
        }

        // Step 3: Compute new Means
        for (i = 0; i < gaussians; i++)
        {
            if (likelihood[i] == 0)
                mean[i] = old_mean[i];
            else
                mean[i] /= likelihood[i];
        }

        // Step 4: Cluster Quality estimation
        for (row = 0; row < datacount; row++)
            quality[featclass[row]] +=
                euclidean((*data)[row], mean[featclass[row]]);
        int changed = 0;
        for (i = 0; i < gaussians; i++)
        {
            quality[i] = likelihood[i] * likelihood[i];
            if (quality[i] != old_quality[i])
                changed = 1;
        }
        if (changed == 0)
            break;
    }

    // Update the covariance
    for (row = 0; row < datacount; row++)
    {
        cov[featclass[row]] += ((*data)[row] - mean[featclass[row]]) *
            ((*data)[row] - mean[featclass[row]]);
    }
    for (i = 0; i < gaussians; i++)
    {
        cov[i] /= likelihood[i];

        // replace small entries with 1.0
        for (j = 0; j < gaussiandim; j++)
        {
            if (cov[i][j] < MIN_COVARIANCE)
                cov[i][j] += GMM_WIDTH;
        }
    }

    for (i = 0; i < gaussians; i++)
    {
        if (likelihood[i] == 0)
            likelihood[i] = 1;
        prior[i] = (double)(likelihood[i]) / (double)(likelihood.sum());
    }
}

void
moc::clusterby_em(int em)
{
    double normal = pow (2.0*M_PI, gaussiandim/2.0);
    double e, eold = 0.0;
    vec s(gaussians);
    vec diff(gaussiandim);
    int row, i, j, iter;
    vec new_prior(gaussians);
    mat init_cov(gaussians);
    mat post(datacount);
    vec prob(datacount);

    // save covariances to restore if they collapse during em
    for (i = 0; i < gaussians; i++)
    {
        init_cov[i].resize(gaussiandim);
        init_cov[i] = cov[i];
    }
    for (row = 0; row < datacount; row++)
        post[row].resize(gaussians);

    for (iter = 0; iter < em; iter++)
    {
        // Compute the current probability P(X|J) of the data X conditioned on
        // each comnponent density
        s = 1.0;
        for (i = 0; i < gaussians; i++)
        {
            for (j = 0; j < gaussiandim; j++)
                s[i] *= sqrt(cov[i][j]);
            s[i] *= normal;
        }
        for (row = 0; row < datacount; row++)
        {
            for (i = 0; i < gaussians; i++)
            {
                // density function of a multivariate gaussian distribution
                // with a diagonal covariance matrix
                diff = (*data)[row] - mean[i];
                diff *= diff;
                diff /= cov[i];
                post[row][i] = exp(-0.5*diff.sum());
                post[row][i] /= s[i];

                // Compute the posteriors. The probability of each component
                // conditioned on the data P(J|X).
                post[row][i] *= prior[i];
            }

            prob[row] = post[row].sum();
            if (prob[row] != 0.0)
                post[row] /= prob[row];
            else
            {
                prob[row] = 1.0;
                cout << "Help" << endl;
            }
        }

        prob = log(prob);
        e = -1 * prob.sum();
        if ((iter > 1) && (abs(e-eold) < 1.0E-11))
            return;
        eold = e;

        // Adjust new priors
        new_prior = 0.0;
        for (i = 0; i < gaussians; i++)
        {
            for (row = 0; row < datacount; row++)
                new_prior[i] += post[row][i];

        }
        prior = new_prior;
        prior /= datacount;

        // New means
        for (i = 0; i < gaussians; i++)
        {
            for (j = 0; j < gaussiandim; j++)
            {
                mean[i][j] = 0.0;
                for (row = 0; row < datacount; row++)
                    mean[i][j] += post[row][i] * (*data)[row][j];

            }
        }
        for (i = 0; i < gaussians; i++)
            mean[i] /= new_prior[i];

        // Compute new covariances
        for (i = 0; i < gaussians; i++)
            cov[i] = 0.0;

        for (row = 0; row < datacount; row++)
        {
            for (i = 0; i < gaussians; i++)
            {
                diff = (*data)[row] - mean[i];
                diff *= diff;
                diff *= post[row][i];
                cov[i] += diff;
            }
        }
        for (i = 0; i < gaussians; i++)
        {
            cov[i] /= new_prior[i];
            if (cov[i].min() < MIN_COVARIANCE)
                cov[i] = init_cov[i];
        }
    }
}

void
moc::add_gmm(gmm *g)
{
    mdb[g->get_md5sum()] = g;
    g->save_model(db + g->get_md5sum());
}

gmm*
moc::get_gmm_md5(const string &md5)
{
    return mdb[md5];
}

class cmp_gmm
{
    public:
        int operator() (const pair<gmm*, double> &g1,
                const pair<gmm*, double> &g2) const
        {
            // 1,2,3,4
            return g1.second < g2.second;
            // 4,3,2,1
//            return g1.second > g2.second;
        }
};

double
moc::get_progress()
{
    return progr;
}

void
moc::get_nclosest_cancel()
{
    cancel = 1;
}

vector<pair<gmm*, double> >*
moc::get_nclosest(gmm *g, int count, int datacount)
{
    vector<pair<gmm*, double> >* pl = new vector<pair<gmm*, double> >;
    typedef map<string, gmm*>::const_iterator ci;

    progr = 0.05;
    double step = 0.90/mdb.size();

    for (ci p = mdb.begin(); p != mdb.end(); p++)
    {
        pl->push_back(make_pair(p->second,
                    g->get_distance(p->second, datacount)));
        progr += step;

        if (cancel)
        {
            delete pl;
            return NULL;
        }
    }

    sort(pl->begin(), pl->end(), cmp_gmm());
    if ((int)pl->size() > count)
        pl->resize(count);

    progr = 1.0;
    cancel = 0;

    return pl;
}
